/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                        PE.C     (PHDM's EDITOR)                         //
//                                                                         //
// Copyright (C) 2000-2014  Philippe Demeester                             //
// Bugs report: phdm@perso.be                                              //
// Web site   : http://phdm.webay.be                                       //
//                                                                         //
// This program is free software; you can redistribute it and/or           //
// modify it under the terms of the GNU General Public                     //
// License as published by the Free Software Foundation; either            //
// version 2.1 of the License, or (at your option) any later version.      //
//                                                                         //
// This program is distributed in the hope that it will be useful,         //
// but WITHOUT ANY WARRANTY; without even the implied warranty of          //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       //
// Library General Public License for more details or write to the         //
// Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

/**************************************************************************** 
Version 0.5 08 09 2004
  Supports TAB
  Supports LANG i18n
  Checks the display size
  Improving critial situation handling (info) 

Version 0.8 June 2008
  Auto resize
  Line width 256
  UTF8 support for iso 8859-1(5)
	File name auto complete & check
	Overwrite warning (ask)
	Modified warning (ask)

version 1.0 nov 2013
  UTF8 on 2 bytes decode & encode
  Buffer become int instead of char
  Remove Glib calls and string completion
  * 
Version 1.1 August 2014
  UTF-8 fully supported
  * 
*****************************************************************************/


#include "pe.h"

/* utf8 convertion 2 bytes only.

 * Masks          7  C | 3   F
 bytes | bits | representation
     1 |    7 | 0vvvvvvv
     2 |   11 | 110vvvvv 10vvvvvv
     3 |   16 | 1110vvvv 10vvvvvv 10vvvvvv
     4 |   21 | 11110vvv 10vvvvvv 10vvvvvv 10vvvvvv
 */

int from_utf8(int h_ch)
{
int ch=0, next_ch;
utf8_valid=0;
if (h_ch <128)return (h_ch); // 1 byte ...
// chk if really 2 bytes encoded
//ch = h_ch & 0xE0;
//if (ch!=0xC0) return (h_ch); // leave as is ...
if (h_ch >127){ //extract hight value
	if ((h_ch & 0xE0) == 0xC0){ // 110vvvvv
		ch = h_ch & 0x1F; // 5 low bites mask
		ch = ch << 6; // shift left 6
		next_ch = getch();
		next_ch = next_ch & 0x3F;
		utf8_valid=1;
		//printf("\n%u\n",ch | next_ch);
		return (ch | next_ch);
		};
	if ((h_ch & 0xF0) == 0xE0){ // 1110vvvv
		ch = h_ch & 0x0F; // 4 low bites mask
		ch = ch << 12; // shift left 12
		next_ch = getch();
		next_ch = next_ch & 0x3F;
		next_ch = next_ch << 6;
		ch = ch | next_ch;
		next_ch = getch();
		next_ch = next_ch & 0x3F;
		ch = ch | next_ch;
		utf8_valid=1;
		//printf("\n1-%u\n",ch);
		return (ch);
		};
	if ((h_ch & 0xF8) == 0xF0){ // 11110vvv
		ch = h_ch & 0x07; // 3 low bites mask
		ch = ch << 18; // shift left 18
		next_ch = getch();
		next_ch = next_ch & 0x3F;
		next_ch = next_ch << 12;
		ch = ch | next_ch;
		next_ch = getch();
		next_ch = next_ch & 0x3F;
		next_ch = next_ch << 6;
		ch = ch | next_ch;
		next_ch = getch();
		next_ch = next_ch & 0x3F;
		ch = ch | next_ch;
		utf8_valid=1;
		//printf("\n2-%u\n",ch);
		return (ch);
		};
	};

return (h_ch); // some function key...
}


/* The main Kb input loop */
void get_kb()
{
int ch;

ch=getch();
chr=ch;

if (xm<80){
  bip();
  return;
	};
if (utf_8==1) 
	ch=from_utf8(ch);

switch (ch){
  case KEY_ESC : more_in();return;break;
  case KEY_END : end_of_line();break;
  case KEY_C1 :  end_of_line();break;
  case 262 :     start_of_line();break;
  case KEY_A1 :  start_of_line();break;
  case 263 :     back_spc(); modified=1; break;
  case KEY_UP :  go_up(0);break;
  case KEY_DOWN :  go_down(0);break;
  case KEY_RIGHT : go_right();break;
  case KEY_LEFT :  go_left();break;
  case 338 :     page_down();break;
  case KEY_C3 : page_down();break;
  case 339 :    page_up();break;
  case KEY_A3 : page_up();break;
  case KEY_DC : delete(); modified=1; break;
  case KEY_F0+1 : load(1);break;
  case KEY_F0+2 : save();break;
  case KEY_F0+3 : block(0);break;
  case KEY_F0+4 : block(1);break;
  case KEY_F0+5 : del_blk(0); modified=1;break;
  case KEY_F0+6 : find();break;
  case KEY_F0+7 : print();break;
  case KEY_F0+8 : quit_flg=1;break;
  case KEY_F0+9 : load(1);break; 
  case KEY_F0+10 : end();break;
  case KEY_F0+11 : del_line(); modified=1;break;
  case KEY_F0+13 : ten_pages_up();break;
  case KEY_F0+14 : ten_pages_down();break;  
  case 391 : home();break;
  case 386 : end();break;
  case 331 : go_last();break;
  case KEY_RESIZE : resize_terminal();
  };

if (utf8_valid)c=ch;	
else if (ch<256)c=ch; 
else c=0;
switch (c){
  case 0:    break;
  case (10): new_line(); modified=1;break;
//  case (127):back_spc();break;
  case (KEY_TAB):  tab(); modified=1;break;
  default :  insert(); modified=1;
  };
}

void more_in()
{
int ch;
ch=getch();
if (ch==79){
  ch=getch();
  switch (ch){
    case 116 : go_left();break;    
    case 118 : go_right();break;    
    case 120 : go_up(0);break;    
    };
  };
}


/* back_spc */
void back_spc()
{
int i,j,k;
if (curr_pos>0)
  {
  curr_pos--;
  i=curr_pos;
  while (txt[curr_line][i]){
     txt[curr_line][i]=txt[curr_line][i+1];
     i++;
     };
  set_xshift();
  displ_line();
  reloc();
  }
else{ 
/* if the line fits in the upper one do it */
   if (curr_line>0){
    i=eof_line();
    curr_line--;
    j=eof_line();
    if ((i + j) < W-1){
       k=0;
       while (k<i){
         txt[curr_line][j]=txt[curr_line+1][k];
         k++;
         j++; 
         };
       shift_txt_up();
       curr_pos=j-i;
       if(curs_y>0){
          curs_y--;
          set_xshift();
          redo_full_scr();
          //displ_line();
          put_line_num();
          reloc();
         }
       else {
        set_xshift();
        redo_full_scr();
        //displ_line();
        put_line_num();
        reloc();
        };
      }
      else {
        bip();
        curr_line++;
      }; 
    }else bip();
  };
last_inst_x=curr_pos;
last_inst_y=curr_line;
put_line_num();
reloc();
}

/* delete */
void delete()
{int i;
if (txt[curr_line][curr_pos]!=0){
i=curr_pos;
while (txt[curr_line][i]){
  txt[curr_line][i]=txt[curr_line][i+1];
  i++;
  };
set_xshift();
displ_line();
put_line_num();
reloc();
last_inst_x=curr_pos;
last_inst_y=curr_line;
del_blk_flg = 1;
}
else bip();
}



/* TAB */
void tab()
{
//c=187;
insert();
}



/* escape chr */
void escape() // never used !!
{
c=getch();
if (c==91){
  c=getch();
  if (c==68){go_left();return;};
  if (c==67){go_right();return;};
  if (c==65){go_up(0);return;};
  if (c==66){go_down(0);return;};
  if (c==51){delete();c=getch();return;};
  if (c==53){page_up();c=getch();return;};
  if (c==54){page_down();c=getch();return;};
  if (c==72){home();return;};
  if (c==70){end();return;};
  if (c==49){
    c=getch();
    if (c==56){quit_flg=1;c=getch();return;}; /* F7 */
    if (c==49){c=getch();load(1);return;};  /* F9 */
    if (c==50){c=getch();save();return;};  /* F2 */
    if (c==51){block(0);c=getch();return;}; /* F3 */
    if (c==53){del_line();c=getch();return;}; /* F5 */
    };
  };
}

/* delete a line */
void del_line()
{
cl_txt_line();
int_strcpy(txt[curr_line],txt[curr_line + 1]);
shift_txt_up();  
curr_pos=0;
if (curr_line>last_line){
  curr_line=last_line;
  curs_y--;
  };
set_xshift();
displ_line();
put_line_num();
reloc();
del_blk_flg = 1;
}

/* delete the marked block */
void del_blk()
{
int i;
i=last_blk_line - first_blk_line;
i++;
if ((blk_flg==2)&&(del_blk_flg==0)){
  curs_y=0;
  curr_line=first_blk_line;
  while( i>0){
    curs_y=0;
    del_line();
    i--;
    };
	del_blk_flg = 1;
	put_blk_status();
	last_inst_x = 0;
	last_inst_y = curr_line;
	go_last();
  }
else bip();
}

/* blocks treatment */
void block(int flag)
{
int i,j,k;
switch (blk_flg){
  case 0:
  if (flag==0){
    blk_flg=1;
    put_blk_status();
    tit();
    first_blk_line=curr_line;
    }
  else bip();
  break;

  case 1:
  if (flag==0){
    if ((curr_line>=first_blk_line)&&(curr_line - first_blk_line < BLK-12)){
      blk_flg = 2;
      del_blk_flg = 0;
      put_blk_status(); 
      last_blk_line=curr_line;
      titat();
  /* cpy marked text in buffer */
      i=0;
      while (first_blk_line+i <= last_blk_line){
        k=0;
        while (k < W){
          block_buf[i][k] = txt[first_blk_line + i][k];
          k++;
          };
        i++;
        };
      }
    else {
      blk_flg=first_blk_line=last_blk_line=0;
      put_blk_status();
      titatut();
      bip();bip();
      info(gettext("Too many lines in block!"));
      };
    }
  else bip();
  break;

  case 2:
  if (flag==1){/* INSERT CASE shift txt n line down */
    j=last_blk_line - first_blk_line + 1;
    i=last_line;
    if (i+j>BUF-64){
      bip();bip();
      info(gettext("The block can't be inserted, buffer overflow!"));
      return;
      };
    while (i > curr_line){
      k=0;
      while(k<W){txt[i+j][k]=txt[i][k];k++;};
      i--;
     };
    last_line=last_line+j;
/* insert txt */
    i=0;
    while (i<j){
      k=0;
      while (k<W){
        txt[curr_line+i+1][k]=block_buf[i][k];
        k++;
        };
      i++;
      };
    i=last_blk_line-first_blk_line;
    i++;
    curr_line=curr_line+i;
    curr_pos=eof_line();
    last_inst_y=curr_line;
    last_inst_x=curr_pos;
    go_last();
    put_cmd();
    reloc();
    del_blk_flg = 1;
    };
  if (flag==0){
    blk_flg = first_blk_line = last_blk_line = 0;
    titatut();
    put_cmd();
    reloc();
    };
  break;
  };
}

/* page up
 * cursor goes on start of line */
void page_up()
{
if (curr_line == 0){
  bip();
  return;
  };
if (curr_line > y_max+1){
  curr_pos = curs_y = xshift = 0;
  curr_line = curr_line - (y_max+1);
  redo_full_scr();
  put_line_num();
  reloc();
  }
else {
  curr_pos = curs_y = xshift = 0;
  curr_line=0;
  redo_full_scr();
  put_line_num();
  reloc();
  };
}

/* ten pages up */
void ten_pages_up()
{
int i=0;
for (i=0;i<10;i++)page_up();
}


/* page down
 * cursor goes on start of line */
void page_down()
{
if (curr_line==last_line){
  bip();
  return;
  };
if (curr_line < last_line - (y_max+1)){
  curr_pos = curs_y = xshift = 0;
  curr_line= curr_line + (y_max+1);
  redo_full_scr();
  put_line_num();
  reloc();
  }
else {
  curr_pos = curs_y = xshift = 0;
  curr_line= last_line;
  redo_full_scr();
  put_line_num();
  reloc();
  };
}

/* ten pages down */
void ten_pages_down()
{
int i=0;
for (i=0;i<10;i++)page_down();
}

/* return to top of txt */
void home()
{
curr_pos=curs_y=curr_line=0;
set_xshift();
redo_full_scr();
put_line_num();
reloc();
}

/* Home Key !!! */
void start_of_line()
{
curr_pos= 0;
set_xshift();
displ_line();
put_line_num();
reloc();
}
/* End Key */
void end_of_line()
{
curr_pos=eof_line();
set_xshift();
displ_line();
put_line_num();
reloc();
}

/* go to end of txt F10 */
void end()
{
if (last_line>y_max){
  curr_line=last_line-10;
  curs_y=0;
//  redo_scr();
  curs_y=10;
  curr_line=last_line;
  curr_pos=eof_line();
  set_xshift();
  redo_full_scr();
  put_line_num();
  reloc();
  }
else {
  curr_line=0;
  curs_y=0;
//  redo_scr();
  curs_y=last_line;
  curr_line=last_line;
  curr_pos=eof_line();
  set_xshift();
  redo_full_scr();
  put_line_num();
  reloc();
  };
}

/* arrow */
void go_left()
{
if (curr_pos>0){
  curr_pos--;
  set_xshift();
  displ_line();
  put_line_num();
  reloc();
   }
else {
  tatit();
  go_up(1);
  };
}

void go_right()
{
if (txt[curr_line][curr_pos]!=0){
  curr_pos++;
  set_xshift();
  displ_line();
  put_line_num();
  reloc();
  }
else {
  tatit();
  go_down(1);
  };
}

/*
 * go_up() & go_down() moves cursor up & down
 * if flag force at eof_line_pos.
 * curr_pos is the txt array pos. but not the curs_x val because of TABs...
 * First find the curs_x in the orig. line (orig_x), use set_curs_pos(0).
 * Then mov to target line and find the curr_pos for the orig_x value
 * if too far, set it to eof_line
 */ 
void go_up(int flag)
{
int orig_x, // original curs pos
	lcurr_pos; // target curs pos
if (curr_line > 0){
  // where is the orig x cursor?  
  orig_x = set_curs_pos(0);
  if (curs_y == 0){
    curr_line--;
    // whats the curr_pos in the upper line?
    lcurr_pos=get_curr_pos(orig_x);
    if (flag==1)lcurr_pos=eof_line();
    curr_pos=lcurr_pos;
    scroll_down();
    set_xshift();
    displ_line();
    put_cmd();
    put_line_num();
    reloc();
    }
  else{
   curr_line--;
   curs_y--;
   // whats the curr_pos in the upper line?
   lcurr_pos=get_curr_pos(orig_x);
   //printf("\n\n\n%u %u %u\n",eof_line(),orig_x,lcurr_pos);
   if (flag==1)lcurr_pos=eof_line();
   curr_pos=lcurr_pos;
   set_xshift();
   displ_line();
   put_line_num();
   reloc();
   };
  }
else bip();
}

void go_down(int flag)
{int orig_x, // original curs pos
	lcurr_pos; //
if (curr_line < last_line){
  orig_x = set_curs_pos(0);	
  if (curs_y == y_max){
    curr_line++;
    // whats the curr_pos in the upper line?
    lcurr_pos=get_curr_pos(orig_x);
    if (lcurr_pos>eof_line())lcurr_pos=eof_line();
    if (flag==1)lcurr_pos=0;
    curr_pos=lcurr_pos;
    set_xshift();
    scroll_up();
    displ_line();
    put_cmd();
    put_line_num();
    reloc();
    }
  else{
    curs_y++;
    curr_line++;
    // whats the curr_pos in the upper line ?
    lcurr_pos=get_curr_pos(orig_x);
    if (lcurr_pos>eof_line())lcurr_pos=eof_line();
    if (flag==1)lcurr_pos=0;
    curr_pos=lcurr_pos;
    set_xshift();
    displ_line();
    put_line_num();
    reloc();
    };
  }
else bip();
}

/* return to the last change */
void go_last()
{
if ((last_inst_y==0)&&(last_inst_x==0)){
  bip();
  return;
  };
if (last_inst_y<12){
  curs_y=0;  
  curr_line=0;
  redo_scr();
  curs_y=last_inst_y;
  curr_pos=last_inst_x;
  curr_line=last_inst_y;
  set_xshift();
  displ_line();
  reloc();
  put_line_num();
  }
else{
  curr_line=last_inst_y-12;
  curs_y=0;  
  redo_scr();
  curs_y=12;
  curr_pos=last_inst_x;
  curr_line=last_inst_y;
  set_xshift();
  displ_line();
  reloc();
  put_line_num();
  };
}

/* insert a new line */
void new_line()
{
/* chk BUF */
if (last_line > BUF-64){
  bip();bip();
  info("End of buffer has been reached!");
  return;
  };
//xshift=0;
/* if we are at last txt line*/
if (curr_line==last_line){
  if (curs_y == y_max){
    move_eol();
    curr_line++;
    curr_pos=0;
    set_xshift();
    scroll_up();
    put_cmd();
    put_line_num();
    displ_line();
    reloc();
    }
  else{
    move_eol();
    put_line();
    curs_y++;
    curr_line++;
    curr_pos=0;
    put_line_num();
    set_xshift();
    displ_line();
    reloc();
    };
  last_line++;
  }
else {
  curr_line++;
  shift_txt_down();
  cl_txt_line();
  curr_line--;
  move_eol();
  set_xshift();
  displ_line();
  curr_line++; 
  if (curs_y == y_max){
    curr_pos=0;
    set_xshift();
    scroll_up();
    displ_line();
    put_cmd();
    put_line_num();
    reloc();
    }
  else{
    curs_y++;
    curr_pos=0;
    set_xshift();
    displ_line();  
    redo_scr();
    put_line_num();
    reloc();
    };
  last_line++;
  };
last_inst_x=curr_pos;
last_inst_y=curr_line;
del_blk_flg = 1;
}

/* move the txt on the right to next line */
void move_eol()
{
int i=0;
while (txt[curr_line][curr_pos + i]){
  txt[curr_line + 1][i] = txt[curr_line][curr_pos + i];
  txt[curr_line][curr_pos + i] = 0;
  i++;
  };
}
   
/*
void pscroll_part_down()
{
int i;
char empty[x_max+1];
i=0;
while (i<(x_max+1)){empty[i]=' ';i++;};
empty[i]=0;
i=0;
while ((curs_y + i)< y_max){
  loc(curs_y + i,0);
  printw("%s",empty);
  loc(curs_y + i,0);
  printw("%s",txt[curr_line + i]);
  i++;};
refresh(); 
}
*/
void shift_txt_down()
{
int i,j;
i=last_line;
while (i >= curr_line){
  j=0;
  while((j<W)&&(txt[i][j]!=0)){txt[i+1][j]=txt[i][j];j++;};
  for(j=0;j<W;j++){txt[i][j]=0;};// clean the line
  i--;
  };
/*clear current line*/
i++;
j=0;
while(j<W){txt[i][j]=0;j++;};
}

void shift_txt_up()
/* move the text up from the curr_line + 1 */
{
int i,j;
i=curr_line+1;
while (i <= last_line+1){
  for(j=0;j<W;j++){txt[i][j]=0;}; // clean first to be sure
  j=0;
  while((j<W)&(txt[i+1][j]!=0)){txt[i][j]=txt[i+1][j]; j++;};
  i++;
  };
last_line--;
}

/* move the txt up */
void scroll_up()
{
int i, tmp_curs_y, tmp_curr_line;
tmp_curs_y = curs_y;
cls();
i = curr_line - y_max;
curs_y = 0;
while (i<= curr_line){
/*
  loc(y,0);
  printw("%s",txt[i]);
 */ 
  tmp_curr_line=curr_line;
  curr_line = i;
  put_line();
  curr_line=tmp_curr_line;
  i++;
  curs_y++;
  };
curs_y = tmp_curs_y;
}

/* move the txt down 
!the curr_line should be adjusted before calling */
void scroll_down()
{
int i,tmp_curs_y, tmp_curr_line;
tmp_curs_y = curs_y;
cls();
i = curr_line;
curs_y = 0;
while (i<= curr_line + y_max){
/*
  loc(y,0);
  printw("%s",txt[i]);
 */
  tmp_curr_line=curr_line;
  curr_line = i;
  put_line();
  curr_line=tmp_curr_line;
  i++;
  curs_y++;
  };
curs_y = tmp_curs_y;
}

/* proc to clear curr_line */
void cl_txt_line()
{
int i;
i=0;
while (i<W){txt[curr_line][i]=0; i++;};
}
/************************** load a file in buffer *************************/
int from_utf8_file(unsigned int h_ch, FILE *f)
{
unsigned int ch=0, next_ch;
if (h_ch == EOF)return (h_ch); // 
if (h_ch <128)return (h_ch); // 1 byte ...
// chk if really 2 bytes encoded

if (h_ch >127){ //extract hight value
	if ((h_ch & 0xE0) == 0xC0){ // 110vvvvv
		ch = h_ch & 0x1F; // 5 low bites mask
		ch = ch << 6; // shift left 6
		next_ch = fgetc(f);
		next_ch = next_ch & 0x3F;
		utf8_valid=1;
		//printf("\n%u\n",ch | next_ch);
		return (ch | next_ch);
		};
	if ((h_ch & 0xF0) == 0xE0){ // 1110vvvv
		ch = h_ch & 0x0F; // 4 low bites mask
		ch = ch << 12; // shift left 12
		next_ch = fgetc(f);
		next_ch = next_ch & 0x3F;
		next_ch = next_ch << 6;
		ch = ch | next_ch;
		next_ch = fgetc(f);
		next_ch = next_ch & 0x3F;
		ch = ch | next_ch;
		utf8_valid=1;
		//printf("\n1-%u\n",ch);
		return (ch);
		};
	if ((h_ch & 0xF8) == 0xF0){ // 11110vvv
		ch = h_ch & 0x07; // 3 low bites mask
		ch = ch << 18; // shift left 18
		next_ch = fgetc(f);
		next_ch = next_ch & 0x3F;
		next_ch = next_ch << 12;
		ch = ch | next_ch;
		next_ch = fgetc(f);
		next_ch = next_ch & 0x3F;
		next_ch = next_ch << 6;
		ch = ch | next_ch;
		next_ch = fgetc(f);
		next_ch = next_ch & 0x3F;
		ch = ch | next_ch;
		utf8_valid=1;
		//printf("\n2-%u\n",ch);
		return (ch);
		};
	};
return (h_ch);
}
void load(int flg)
{
int i, j, c=0, cr=0;

char backup_file[256], lfile_name[256];
FILE *f, *bu_f; /* backup file */

i=0;
backup_file[0]=0;
if(flg==0){ // build full path
	if(file_name[0]!='/'){
		i=strlen(file_name);
		while(i>0){
			if (file_name[i]=='/')break;
			i--;
			};
		for (j=0;j<i;j++)backup_file[j]=file_name[j];
		backup_file[j]=0;
		chdir(backup_file);
		getcwd(lfile_name,127);
		if (is_this_dir(lfile_name)==0){
			if (strlen(lfile_name)<126)
				strcat(lfile_name, "/");
			else{
				info(gettext("File path too long."));
				return;
				};
			strcpy(curr_dir,lfile_name);
			};
		if (file_name[j]=='/') j++;
		if ((strlen(lfile_name)+strlen(&file_name[j]))<126){
			strcat(lfile_name,&file_name[j]);
			strcpy(file_name, lfile_name);
			}
		else{
			info(gettext("File path too long."));
			return;
			};
		};
	};
if ((file_name[0]!=0)&&(flg==1)){
  bip();bip();
  info(gettext("A file is already loaded!"));
  return;
  }
else if (txt[0][0]!=0){
  bip();bip();
  info(gettext("The buffer is not empty!"));
  return;
  }
else{
  if (flg==1){
    color(2);
    curs_y= y_max+1;
    cll();
    loc( y_max+1,1);
/*
    printw("%s",gettext("File:"));
    echo(); 
    getnstr(file_name,70);
    noecho();
 */ 
	get_fname(0);
    };
  f=fopen(file_name,"rb");
  if (f==NULL){
    bip(),bip(),bip();/* no such file */
    info (gettext("File not found!"));
    file_name[0]=0;
    }
  else {
    fclose(f);
    /* check if writable */
    f=fopen(file_name,"a");
    if (f!=NULL){
      fclose(f);
      f=fopen(file_name,"rb");
      /* make a backup */
      sprintf(backup_file,"%s~",file_name);
      bu_f=fopen(backup_file,"wb");
      if (bu_f==NULL){
        bip(),bip();
        color(2);
        cll();
        bip();bip();
        info(gettext("Can't create the backup file"));
        }
      else{
        while (c!=EOF){
          c=fgetc(f);
          if (c!=EOF)fputc(c,bu_f);
					};
				fclose(bu_f);
				rewind(f);
				};
      }
    else{ /* the file is read only */
      bip(),bip();
      color(2);
      cll();
      bip();bip();
      info(gettext("This file is Read Only!"));
      f=fopen(file_name,"rb");
      };
     i=j=0;
     c=0;
    while (c!=EOF){
      c=fgetc(f);
	/* check utf8 */
	if (utf_8==1){ 
		c=from_utf8_file(c, f);
		};
    /* if we reach EOL put '\' and go to next line */
      curr_line=i;curr_pos=j;
      if(c!=10)
        if (j==W-2) {
          txt[i][j]='\\';
          cr=c;
          c=10;
          };
      if (c==10){
        if (txt[i][j-1]=='\\')txt[i][j]='\\';/* add \ if end of line */
        i++;j=0;
        if (cr!=0){txt[i][j]=cr;
          cr=0;
          j++;
         };
        if (i> BUF-64){
          fclose(f);
          bip();bip();
          info(gettext("This file has too many lines to fit in PE."));
          last_line=i;
          return;
          }; 
        }
      else {
        if (c!=EOF)txt[i][j]=c;
        j++;  
        };
      };
      fclose(f);
      last_line=i;
    };
  put_cmd();
  put_line_num();
  color(1);
  curs_y=0;curs_x=0;
  curr_line=curr_pos=xshift=0; 
  redo_scr();
  reloc();
  };
}

/*************************** find a string in txt *************************/
void from_utf8_str(char *str)
{
int i,j;
unsigned int ch=0, next_ch,h_ch;
i=j=ch=0;
while(str[i]!=0){
	ch=0;
	h_ch=str[i];
	if (h_ch <128){
			 find_str[j]=(h_ch); // 1 byte ...
			 i++;j++;
			};
	if (h_ch >127){ //extract hight value
		if ((h_ch & 0xE0) == 0xC0){ // 110vvvvv
			ch = h_ch & 0x1F; // 5 low bites mask
			ch = ch << 6; // shift left 6
			i++;
			next_ch = str[i];
			next_ch = next_ch & 0x3F;
			//printf("\n%u\n",ch | next_ch);
			find_str[j]=(ch | next_ch);
			i++;j++;
			};
		if ((h_ch & 0xF0) == 0xE0){ // 1110vvvv
			ch = h_ch & 0x0F; // 4 low bites mask
			ch = ch << 12; // shift left 12
			i++;
			next_ch = str[i];
			next_ch = next_ch & 0x3F;
			next_ch = next_ch << 6;
			ch = ch | next_ch;
			i++;
			next_ch = str[i];
			next_ch = next_ch & 0x3F;
			ch = ch | next_ch;
			//printf("\n1-%u\n",ch);
			find_str[j]=ch;
			i++;j++;
			};
		if ((h_ch & 0xF8) == 0xF0){ // 11110vvv
			ch = h_ch & 0x07; // 3 low bites mask
			ch = ch << 18; // shift left 18
			i++;
			next_ch = str[i];
			next_ch = next_ch & 0x3F;
			next_ch = next_ch << 12;
			ch = ch | next_ch;
			i++;
			next_ch = str[i];
			next_ch = next_ch & 0x3F;
			next_ch = next_ch << 6;
			ch = ch | next_ch;
			i++;
			next_ch = str[i];
			next_ch = next_ch & 0x3F;
			ch = ch | next_ch;
			//printf("\n2-%u\n",ch);
			find_str[j]=ch;
			i++;j++;
			};
		};
	};
find_str[j] = 0;
//printf("\n2-%s\n",find_str);
}

void find()
{
int i, l, sl;
char c;
static char lfind_str[160];
reprompt:;
color(2);
i=curs_y;
curs_y= y_max+1;
cll();
loc( y_max+1,0);
curs_y=i; 
if (lfind_str[0]==0){
  printw("%s",gettext("Find: "));
  echo();
  getnstr(lfind_str,70);
  noecho();
  }
else {
  printw("%s %s",gettext("Find: "),lfind_str);
  c=getch();
  if (c!=10){
    lfind_str[0]=0;
    goto reprompt;
    };
  };
l=strlen(lfind_str);
i=curr_line;

if (l>0){
	from_utf8_str(lfind_str);
	sl=int_strlen(find_str);
	while ((!search_line(i,sl)) && (i<=last_line)){i++;};
	if (i<last_line){
	    curs_y=0;
	    curr_line=i;
    }
  else bip();
  }
else bip();
put_cmd();
put_line_num();
color(1);
set_xshift();
redo_full_scr();
reloc();
//printf("\n%d\n",txt[0][2]);
}

int search_line(int i,int l)
{
int j,k;
j=k=0;
if (i==curr_line)j=curr_pos;
while (txt[i][j]!=0){
  if (txt[i][j]==find_str[k]){j++;k++;}
  else {k=0;j++;};
  if (k==l){
		curr_pos=j;
		return 1;
		}; 
	};
return 0;
}

/**************************  save the buffer in file **********************/
void save()
{
int curs_s,i,j;
char prev_fname[128];
int cu;
FILE *f;
unsigned int ch,lch,hch,tch,fch;
i=0;
curs_s = curs_y;

color(2);
curs_y = y_max+1;
cll();
loc( y_max+1,1);

printw("%s%s",gettext("File:"),file_name);

strcpy(prev_fname,file_name);
get_fname(1); // try completion ... (supressed in v1.0)
/* chk previous name */
if (strcmp(prev_fname,file_name)!=0){
	if (is_this_file(file_name)==0)
		if(ask(gettext("Overiting existing file [y/N]?"))==0){
		strcpy(file_name, prev_fname);
		put_cmd();
		put_line_num();
		color(1);
		curs_y=curs_s;
		reloc();
		return;
		};
	};
/* save the file */
f=fopen(file_name,"wb");
if (f==NULL){
  bip();bip();bip();
  info(gettext("This file can't be saved."));
  }
else{
  i=j=0;
  while (i<=last_line){
    switch(txt[i][j]){
      case (0)   :if(i<last_line){fputc('\n',f);j=0;i++;}
              else i++;break;
      case ('\\'):if (txt[i][j+1]==0){i++;j=0;}
                 else if((txt[i][j+1]=='\\')&(txt[i][j+2]==0)){
                    fputc('\\',f);fputc('\n',f);j=0;i++;}
                  else{ fputc(txt[i][j],f);j++;};
                 break;
      default : {
				if (utf_8==1){ 
					cu=txt[i][j];
					if ((cu)<128){
						fputc(txt[i][j],f);
						}
					else if (cu<2048){ // 2 bytes
						ch=cu;
						lch=ch&0x03F;
						lch=lch|0x080; //put b10 in front
						hch=ch&0x7C0;
						hch=hch>>6;
						hch=hch|0xC0;//put b110 in front
						fputc(hch,f);
						fputc(lch,f);
						}
					else if (cu<65536){ // 3 bytes
						ch=cu;
						lch=ch&0x03F;
						lch=lch|0x080; //put b10 in front
						hch=ch&0x0FC0;  // 0000 1111 1100 0000
						hch=hch>>6;
						hch=hch|0x080; //put b10 in front
						tch=ch&0x0F000;  // 1111 0000 0000 0000
						tch=tch>>12;
						tch=tch|0xE0;//put b1110 in front
						fputc(tch,f);
						fputc(hch,f);
						fputc(lch,f);
						}
					else{ // 4 bytes
						ch=cu;
						lch=ch&0x03F;
						lch=lch|0x080; //put b10 in front
						hch=ch&0x0FC0;  // 0000 1111 1100 0000
						hch=hch>>6;
						hch=hch|0x080; //put b10 in front
						tch=ch&0x3F000;  //  0011 1111 0000 0000 0000
						tch=tch>>12;
						tch=tch|0x080; //put b10 in front
						fch=ch&0x01C0000;  // 0001 1100 0000 0000 0000 0000
						fch=fch>>18;
						fch=fch|0xF0;//put b11110 in front
						fputc(fch,f);
						fputc(tch,f);
						fputc(hch,f);
						fputc(lch,f);		
						};
						j++;
					}
				else{
						fputc(txt[i][j],f);
						j++;
						};
				}; 
				break;
		};
    };
	fclose(f);
	modified=0;
  };
put_cmd();
put_line_num();
color(1);
curs_y=curs_s;
reloc();
}

/***************************** Send to printer **************************/
void print()
{
int i,j;
FILE *f;
unsigned int c,cu,ch,lch,hch,tch,fch; 
/* copy the txt to pe.lpr */
f=fopen(".pe.lpr","w");
// 
if (f==NULL)return;
else{
  i=j=0;
  while (i<=last_line){
    c=txt[i][j];
    if ((c==0)&&(i<last_line)){c=10;fputc(c,f);i++;j=0;} // end of line
    else if ((c==0)&&(i==last_line)){i++;} // this the end...
//  else if (c==-69){c=9;fputc(c,f);j++;} /* convert tab to its real value */
    else { // utf-8?
      if (utf_8==1){ 
		cu=txt[i][j];
		if ((cu)<128){
			fputc(cu,f);
			}
		else if (cu<2048){ // 2 bytes
			ch=cu;
			lch=ch&0x03F;
			lch=lch|0x080; //put b10 in front
			hch=ch&0x7C0;
			hch=hch>>6;
			hch=hch|0xC0;//put b110 in front
			fputc(hch,f);
			fputc(lch,f);
			}
		else if (cu<65536){ // 3 bytes
			ch=cu;
			lch=ch&0x03F;
			lch=lch|0x080; //put b10 in front
			hch=ch&0x0FC0;  // 0000 1111 1100 0000
			hch=hch>>6;
			hch=hch|0x080; //put b10 in front
			tch=ch&0x0F000;  // 1111 0000 0000 0000
			tch=tch>>12;
			tch=tch|0xE0;//put b1110 in front
			fputc(tch,f);
			fputc(hch,f);
			fputc(lch,f);
			}
		else{ // 4 bytes
			ch=cu;
			lch=ch&0x03F;
			lch=lch|0x080; //put b10 in front
			hch=ch&0x0FC0;  // 0000 1111 1100 0000
			hch=hch>>6;
			hch=hch|0x080; //put b10 in front
			tch=ch&0x3F000;  //  0011 1111 0000 0000 0000
			tch=tch>>12;
			tch=tch|0x080; //put b10 in front
			fch=ch&0x01C0000;  // 0001 1100 0000 0000 0000 0000
			fch=fch>>18;
			fch=fch|0xF0;//put b11110 in front
			fputc(fch,f);
			fputc(tch,f);
			fputc(hch,f);
			fputc(lch,f);		
			};
			j++;
		}
	else{			
		fputc(c,f);j++;
		};
    };
  };    
  fclose(f);
/* now print */
  if(system ("pe-prt .pe.lpr")){
    bip();bip();bip();
    info(gettext("Please check or create <pe-prt> script."));
    def_prog_mode();
    endwin();
    refresh();
    reloc();
    };
  }; 
unlink(".pe.lpr");
}

/*********************** prepare to quit pe ****************************/
void quit()
{
/* if modified, ask for save */
if (modified==1)
	if (ask(gettext("File modified, save ? [y/N]:")))save();
attroff(COLOR_PAIR(current_color));
erase();
refresh();
keypad(stdscr,FALSE);
endwin();

}

/*****************************************************************************
* main program entry 
* arg = file name 
*****************************************************************************/
int main(int argc,char *argv[])
{
char *domain, *lang, *home;
int i;
x_max=X;
y_max=Y;
TAB_VAL=8;
comfirm_flg=1;

last_line = curr_line = curr_pos = curs_y = xshift = 0;
//gs=g_string_new("");
//gcmpl=g_completion_new(NULL);
//items=NULL;
home=getenv("HOME");
if (strlen(home)<64)
	strcpy(home_str, home);

utf_8 = utf8_valid = 0;
// locale: check if utf8
lang=getenv("LANG");
if (((strstr(lang,"utf")!=NULL)||(strstr(lang,"UTF")!=NULL))&&
	(strstr(lang,"8")!=NULL))
	utf_8 = 1;
	
if (setlocale(LC_ALL, lang)!=NULL){
	domain=textdomain("pe");
	};
/* // some DEBUG
printf("LANG=%s UTF-8=%u\n",getenv("LC_MESSAGES") , utf_8);
i= getchar();
*/

init();
/*resize signal
if (signal (SIGWINCH, resize_handler) == SIG_IGN)
 puts("signal handler failed!");
*/ 
y_max=ym-2;
if (xm<80){
  bip();
  quit();
  puts(gettext("PE fits in minimum 80 characters width screen...")); 
  exit(0); 
};
if (xm>255){
  bip();
  quit();
  puts(gettext("PE fits in max 256 characters width screen...")); 
  exit(0); 
};
x_max=xm-1;
cls();
intro();
sleep(1);
init_tab_len();
if (argc==2){
  if (strstr(argv[1],"-t")!=NULL) tit_flg=1;
  else{
    tit_flg=0;
		if (strlen(argv[1])<128){
			strcpy(file_name,argv[1]);
			load(0);
			};
    };
  };
if (argc>2){
  if (strstr(argv[1],"-t")!=NULL) tit_flg=1;
  else tit_flg=0;
	if (strlen(argv[2])<128){	
		strcpy(file_name,argv[2]);
		load(0);
		};
  tut();
  };
put_cmd();
put_line_num(); 
if (file_name[0]==0)tut();
/* Debug input
while (quit_flg == 0){
  int ch;
  char str[32];
  ch=getch();
  sprintf(str, "%d ",ch);
  addstr(str);
  refresh();
  };
//  */
/* the loop */
while (quit_flg == 0)get_kb();

quit();
titatut();
exit(0);
}

